@@ -32,15 +32,15 @@ except ImportError:
 
 
 # Initialise the constants.
-PYQT_VERSION_STR = "5.3.1"
+PYQT_VERSION_STR = "5.4"
 
-SIP_MIN_VERSION = '4.16'
+SIP_MIN_VERSION = '4.16.4'
 
 
 class ModuleMetadata:
     """ This class encapsulates the meta-data about a PyQt5 module. """
 
-    def __init__(self, qmake_QT=None, qmake_TARGET='', qpy_lib=False):
+    def __init__(self, qmake_QT=None, qmake_TARGET='', qpy_lib=False, cpp11=False):
         """ Initialise the meta-data. """
 
         # The values to update qmake's QT variable.
@@ -53,6 +53,9 @@ class ModuleMetadata:
         # Set if there is a qpy support library.
         self.qpy_lib = qpy_lib
 
+        # Set if C++11 support is required.
+        self.cpp11 = cpp11
+
 
 # The module meta-data.
 MODULE_METADATA = {
@@ -66,6 +69,7 @@ MODULE_METADATA = {
                                     qpy_lib=True),
     'QtDesigner':           ModuleMetadata(qmake_QT=['designer'],
                                     qpy_lib=True),
+    'Enginio':              ModuleMetadata(qmake_QT=['enginio']),
     'QtGui':                ModuleMetadata(qpy_lib=True),
     'QtHelp':               ModuleMetadata(qmake_QT=['help']),
     'QtMacExtras':          ModuleMetadata(qmake_QT=['macextras']),
@@ -85,12 +89,20 @@ MODULE_METADATA = {
     'QtSql':                ModuleMetadata(qmake_QT=['sql', 'widgets']),
     'QtSvg':                ModuleMetadata(qmake_QT=['svg']),
     'QtTest':               ModuleMetadata(qmake_QT=['testlib', 'widgets']),
+    'QtWebChannel':         ModuleMetadata(qmake_QT=['webchannel', 'network']),
+    'QtWebEngineWidgets':   ModuleMetadata(
+                                    qmake_QT=['webenginewidgets', 'network',
+                                            'widgets'],
+                                    cpp11=True),
     'QtWebKit':             ModuleMetadata(qmake_QT=['webkit', 'network']),
-    'QtWebKitWidgets':      ModuleMetadata(qmake_QT=['webkitwidgets']),
+    'QtWebKitWidgets':      ModuleMetadata(
+                                    qmake_QT=['webkitwidgets',
+                                            'printsupport']),
     'QtWebSockets':         ModuleMetadata(qmake_QT=['websockets']),
     'QtWidgets':            ModuleMetadata(qmake_QT=['widgets']),
     'QtWinExtras':          ModuleMetadata(qmake_QT=['winextras', 'widgets']),
     'QtX11Extras':          ModuleMetadata(qmake_QT=['x11extras']),
+    'QtXml':                ModuleMetadata(qmake_QT=['xml', '-gui']),
     'QtXmlPatterns':        ModuleMetadata(
                                     qmake_QT=['xmlpatterns', '-gui',
                                             'network']),
@@ -129,11 +141,12 @@ MODULE_METADATA = {
 COMPOSITE_COMPONENTS = (
     'QtCore',
     'QtDBus', 'QtGui', 'QtNetwork', 'QtSensors', 'QtSerialPort',
-    'QtMultimedia', 'QtQml', 'QtWebKit', 'QtWidgets', 'QtXmlPatterns',
+    'QtMultimedia', 'QtQml', 'QtWebKit', 'QtWidgets', 'QtXml', 'QtXmlPatterns',
     'QtAxContainer', 'QtDesigner', 'QtHelp', 'QtMultimediaWidgets', 'QtOpenGL',
         'QtPrintSupport', 'QtQuick', 'QtSql', 'QtSvg', 'QtTest',
     'QtWebKitWidgets', 'QtBluetooth', 'QtMacExtras', 'QtPositioning',
-    'QtWinExtras', 'QtX11Extras', 'QtQuickWidgets', 'QtWebSockets'
+        'QtWinExtras', 'QtX11Extras', 'QtQuickWidgets', 'QtWebSockets',
+        'Enginio', 'QtWebChannel', 'QtWebEngineWidgets'
 )
 
 
@@ -490,6 +503,7 @@ class TargetConfiguration:
         self.qmake_variables = []
         self.py_pylib_dir = ''
         self.py_pylib_lib = ''
+        self.py_pyshlib = ''
         self.pydbus_inc_dir = ''
         self.pydbus_module_dir = ''
         self.pyqt_disabled_features = []
@@ -560,6 +574,7 @@ class TargetConfiguration:
                 self.py_pylib_dir)
         self.py_pylib_lib = parser.get(section, 'py_pylib_lib',
                 self.py_pylib_lib)
+        self.py_pyshlib = parser.get(section, 'py_pyshlib', self.py_pyshlib)
 
         self.qt_shared = parser.getboolean(section, 'qt_shared',
                 self.qt_shared)
@@ -691,29 +706,41 @@ int main(int argc, char **argv)
             pylib_lib = 'python%d%d%s' % (py_major, py_minor, debug_suffix)
 
             pylib_dir = self.py_lib_dir
+
+            # Assume Python is a DLL on Windows.
+            pyshlib = pylib_lib
         else:
             abi = getattr(sys, 'abiflags', '')
             pylib_lib = 'python%d.%d%s' % (py_major, py_minor, abi)
+            pylib_dir = pyshlib = ''
 
             # Use distutils to get the additional configuration.
             from distutils.sysconfig import get_config_vars
             ducfg = get_config_vars()
 
-            exec_prefix = ducfg['exec_prefix']
-            multiarch = ducfg.get('MULTIARCH', '')
-            libdir = ducfg['LIBDIR']
+            config_args = ducfg.get('CONFIG_ARGS', '')
 
-            if glob.glob('%s/lib/libpython%d.%d*' % (exec_prefix, py_major, py_minor)):
-                pylib_dir = exec_prefix + '/lib'
-            elif multiarch != '' and glob.glob('%s/lib/%s/libpython%d.%d*' % (exec_prefix, multiarch, py_major, py_minor)):
-                pylib_dir = exec_prefix + '/lib/' + multiarch
-            elif glob.glob('%s/libpython%d.%d*' % (libdir, py_major, py_minor)):
-                pylib_dir = libdir
-            else:
-                pylib_dir = ''
+            dynamic_pylib = '--enable-shared' in config_args
+            if not dynamic_pylib:
+                dynamic_pylib = '--enable-framework' in config_args
+
+            if dynamic_pylib:
+                pyshlib = ducfg.get('LDLIBRARY', '')
+
+                exec_prefix = ducfg['exec_prefix']
+                multiarch = ducfg.get('MULTIARCH', '')
+                libdir = ducfg['LIBDIR']
+
+                if glob.glob('%s/lib/libpython%d.%d*' % (exec_prefix, py_major, py_minor)):
+                    pylib_dir = exec_prefix + '/lib'
+                elif multiarch != '' and glob.glob('%s/lib/%s/libpython%d.%d*' % (exec_prefix, multiarch, py_major, py_minor)):
+                    pylib_dir = exec_prefix + '/lib/' + multiarch
+                elif glob.glob('%s/libpython%d.%d*' % (libdir, py_major, py_minor)):
+                    pylib_dir = libdir
 
         self.py_pylib_dir = pylib_dir
         self.py_pylib_lib = pylib_lib
+        self.py_pyshlib = pyshlib
 
         # Apply sysroot where necessary.
         if self.sysroot != '':
@@ -1203,6 +1230,8 @@ def check_modules(target_config, verbose):
             'new QWebPage()')
     check_module(target_config, verbose, 'QtWidgets', 'qwidget.h',
             'new QWidget()')
+    check_module(target_config, verbose, 'QtXml', 'qdom.h',
+            'new QDomDocument()')
     check_module(target_config, verbose, 'QtXmlPatterns', 'qxmlname.h',
             'new QXmlName()')
 
@@ -1217,7 +1246,6 @@ def check_modules(target_config, verbose):
 
     check_module(target_config, verbose, 'QtDBus', 'qdbusconnection.h',
             'QDBusConnection::systemBus()')
-    check_dbus(target_config, verbose)
 
     if target_config.qt_version >= 0x050100:
         check_5_1_modules(target_config, verbose)
@@ -1228,6 +1256,9 @@ def check_modules(target_config, verbose):
     if target_config.qt_version >= 0x050300:
         check_5_3_modules(target_config, verbose)
 
+    if target_config.qt_version >= 0x050400:
+        check_5_4_modules(target_config, verbose)
+
 
 def check_5_1_modules(target_config, verbose):
     """ Check which modules introduced in Qt v5.1 can be built and update the
@@ -1294,6 +1325,20 @@ def check_5_3_modules(target_config, verbose):
             'new QQuickWidget()')
     check_module(target_config, verbose, 'QtWebSockets', 'qwebsocket.h',
             'new QWebSocket()')
+    check_module(target_config, verbose, 'Enginio', 'enginioclient.h',
+            'new EnginioClient()')
+
+
+def check_5_4_modules(target_config, verbose):
+    """ Check which modules introduced in Qt v5.4 can be built and update the
+    target configuration accordingly.  target_config is the target
+    configuration.  verbose is set if the output is to be displayed.
+    """
+
+    check_module(target_config, verbose, 'QtWebChannel', 'qwebchannel.h',
+            'new QWebChannel()')
+    check_module(target_config, verbose, 'QtWebEngineWidgets',
+            'qwebengineview.h', 'new QWebEngineView()')
 
 
 def generate_makefiles(target_config, verbose, no_timestamp, parts, tracing):
@@ -1330,28 +1375,10 @@ def generate_makefiles(target_config, verbose, no_timestamp, parts, tracing):
             qpy_sources = []
             qpy_headers = []
 
-        if mname == 'QtCore':
-            qpy_sources.insert(0, 'qpycore_post_init.cpp')
-
         generate_sip_module_code(target_config, verbose, no_timestamp, parts,
                 tracing, mname, sip_flags, qpy_sources, qpy_headers)
         subdirs.append(mname)
 
-    # Embed the sip flags.
-    if 'QtCore' in target_config.pyqt_modules:
-        inform("Embedding sip flags...")
-
-        in_f = open(source_path('qpy', 'QtCore', 'qpycore_post_init.cpp.in'))
-        out_f = open_for_writing(
-                os.path.join('QtCore', 'qpycore_post_init.cpp'))
-
-        for line in in_f:
-            line = line.replace('@@PYQT_SIP_FLAGS@@', sip_flags)
-            out_f.write(line)
-
-        in_f.close()
-        out_f.close()
-
     # Generate the composite module.
     qtmod_sipdir = os.path.join('sip', 'Qt')
     mk_clean_dir(qtmod_sipdir)
@@ -1481,9 +1508,9 @@ INSTALLS += qscintilla_api
         mode |= (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
         os.chmod(pyuic_wrapper, mode)
 
-    # Generate the top-level makefile.
-    inform("Generating the top-level Makefile...")
-    run_qmake(target_config, verbose, toplevel_pro)
+    # Generate the makefiles.
+    inform("Generating the Makefiles...")
+    run_qmake(target_config, verbose, toplevel_pro, recursive=True)
 
 
 def generate_plugin_makefile(target_config, verbose, plugin_dir, install_dir, plugin_name):
@@ -1496,9 +1523,9 @@ def generate_plugin_makefile(target_config, verbose, plugin_dir, install_dir, pl
     generated.
     """
 
-    # Check we have an interpreter library.
-    if target_config.py_pylib_dir == '':
-        inform("The %s plugin was disabled because the Python library couldn't be found." % plugin_name)
+    # Check we have a shared interpreter library.
+    if target_config.py_pyshlib == '':
+        inform("The %s plugin was disabled because a dynamic Python library couldn't be found." % plugin_name)
         return False
 
     # Create the qmake project file.
@@ -1506,18 +1533,19 @@ def generate_plugin_makefile(target_config, verbose, plugin_dir, install_dir, pl
 
     sp_plugin_dir = source_path(plugin_dir)
 
-    fin = open(os.path.join(sp_plugin_dir, 'python.pro-in'))
+    fin = open(os.path.join(sp_plugin_dir, '%s.pro-in' % plugin_dir))
     prj = fin.read()
     fin.close()
 
     prj = prj.replace('@QTCONFIG@',
             'debug' if target_config.debug else 'release')
-    prj = prj.replace('@PYINCDIR@', quote(target_config.py_inc_dir))
-    prj = prj.replace('@SIPINCDIR@', quote(target_config.sip_inc_dir))
+    prj = prj.replace('@PYINCDIR@', qmake_quote(target_config.py_inc_dir))
+    prj = prj.replace('@SIPINCDIR@', qmake_quote(target_config.sip_inc_dir))
     prj = prj.replace('@PYLINK@', target_config.get_pylib_link_arguments())
-    prj = prj.replace('@QTPLUGINDIR@', quote(install_dir))
+    prj = prj.replace('@PYSHLIB@', target_config.py_pyshlib)
+    prj = prj.replace('@QTPLUGINDIR@', qmake_quote(install_dir))
 
-    pro_name = os.path.join(plugin_dir, 'python.pro')
+    pro_name = os.path.join(plugin_dir, '%s.pro' % plugin_dir)
 
     mk_dir(plugin_dir)
     fout = open_for_writing(pro_name)
@@ -1533,18 +1561,14 @@ VPATH = %s
 
     fout.close()
 
-    # Create the makefile.
-    inform("Generating the %s plugin Makefile..." % plugin_name)
-
-    run_qmake(target_config, verbose, pro_name)
-
     return True
 
 
 def generate_application_makefile(target_config, verbose, src_dir):
     """ Create the makefile for a QtXml based application.  target_config is
     the target configuration.  verbose is set if the output is to be displayed.
-    src_dir is the name of the directory containing the source files.
+    src_dir is the name of the directory containing the source files.  (It is
+    assumed that it is not a path.)
     """
 
     mk_dir(src_dir)
@@ -1555,7 +1579,8 @@ def generate_application_makefile(target_config, verbose, src_dir):
 
     inform("Generating the .pro file for %s..." % app)
 
-    pro_lines = ['TEMPLATE = app', 'QT -= gui', 'QT += xml']
+    pro_lines = ['TARGET = %s' % app]
+    pro_lines.extend(['TEMPLATE = app', 'QT -= gui', 'QT += xml'])
     pro_lines.append(
             'CONFIG += warn_on %s' %
                     ('debug' if target_config.debug else 'release'))
@@ -1565,6 +1590,9 @@ def generate_application_makefile(target_config, verbose, src_dir):
     elif target_config.py_platform == 'darwin':
         pro_lines.append('CONFIG -= app_bundle')
 
+    # Work around QTBUG-39300.
+    pro_lines.append('CONFIG -= android_install')
+
     pro_lines.append('target.path = %s' % target_config.pyqt_bin_dir)
     pro_lines.append('INSTALLS += target')
 
@@ -1575,15 +1603,12 @@ def generate_application_makefile(target_config, verbose, src_dir):
     pro_lines.extend(pro_sources(sp_src_dir))
     pro_lines.extend(target_config.qmake_variables)
 
-    pro_name = os.path.join(src_dir, app + '.pro')
+    pro_name = os.path.join(src_dir, src_dir + '.pro')
 
     pro = open_for_writing(pro_name)
     pro.write('\n'.join(pro_lines))
     pro.close()
 
-    inform("Generating the Makefile for %s..." % app)
-    run_qmake(target_config, verbose, pro_name)
-
 
 def pro_sources(src_dir, other_headers=None, other_sources=None):
     """ Return the HEADERS and SOURCES variables for a .pro file by
@@ -1751,12 +1776,13 @@ def inform_user(target_config, sip_version):
         inform("PyQt5 will only be usable with signed interpreters.")
 
 
-def run_qmake(target_config, verbose, pro_name, makefile_name='', fatal=True):
+def run_qmake(target_config, verbose, pro_name, makefile_name='', fatal=True, recursive=False):
     """ Run qmake against a .pro file.  target_config is the target
     configuration.  verbose is set if the output is to be displayed.  pro_name
     is the name of the .pro file.  makefile_name is the name of the makefile
     to generate (and defaults to Makefile).  fatal is set if a qmake failure is
     considered a fatal error, otherwise False is returned if qmake fails.
+    recursive is set to use the -recursive flag.
     """
 
     # qmake doesn't behave consistently if it is not run from the directory
@@ -1782,6 +1808,9 @@ def run_qmake(target_config, verbose, pro_name, makefile_name='', fatal=True):
         args.append('-o')
         args.append(makefile_name)
 
+    if recursive:
+        args.append('-recursive')
+
     args.append(pro_file)
 
     run_command(' '.join(args), verbose)
@@ -2113,6 +2142,9 @@ def pro_add_qt_dependencies(target_config, metadata, pro_lines, debug=None):
     pro_lines.append(
             'CONFIG += %s' % ('debug qml_debug' if debug else 'release'))
 
+    if metadata.cpp11:
+        pro_lines.append('CONFIG += c++11')
+
     pro_lines.extend(target_config.qmake_variables)
 
 
@@ -2261,6 +2293,21 @@ def generate_sip_module_code(target_config, verbose, no_timestamp, parts, tracin
     if not os.access(os.path.join(mname, file_check), os.F_OK):
         error("Unable to create the C++ code.")
 
+    # Embed the sip flags.
+    if mname == 'QtCore':
+        inform("Embedding sip flags...")
+
+        in_f = open(source_path('qpy', 'QtCore', 'qpycore_post_init.cpp.in'))
+        out_f = open_for_writing(
+                os.path.join('QtCore', 'qpycore_post_init.cpp'))
+
+        for line in in_f:
+            line = line.replace('@@PYQT_SIP_FLAGS@@', sip_flags)
+            out_f.write(line)
+
+        in_f.close()
+        out_f.close()
+
     # Generate the makefile.
     includepath = libs = ''
     if target_config.vend_enabled:
@@ -2303,6 +2350,9 @@ def generate_module_makefile(target_config, verbose, mname, includepath='', libs
 
     pro_add_qt_dependencies(target_config, metadata, pro_lines)
 
+    # Work around QTBUG-39300.
+    pro_lines.append('CONFIG -= android_install')
+
     if not target_config.static:
         debug_suffix = get_win32_debug_suffix(target_config.debug)
         link = target_config.get_pylib_link_arguments()
@@ -2393,8 +2443,8 @@ macx {
     pro_lines.append('TARGET = %s' % target_name)
 
     if src_dir != mname:
-        pro_lines.append('INCLUDEPATH += %s' % quote(src_dir))
-        pro_lines.append('VPATH = %s' % quote(src_dir))
+        pro_lines.append('INCLUDEPATH += %s' % qmake_quote(src_dir))
+        pro_lines.append('VPATH = %s' % qmake_quote(src_dir))
 
     pro_lines.extend(pro_sources(src_dir, qpy_headers, qpy_sources))
 
@@ -2405,9 +2455,6 @@ macx {
     pro.write('\n')
     pro.close()
 
-    inform("Generating the Makefile for the %s module..." % target_name)
-    run_qmake(target_config, verbose, pro_name)
-
 
 def pro_add_qpy(mname, metadata, pro_lines):
     """ Add the qpy dependencies of a module to a .pro file.  mname is the
@@ -2450,10 +2497,11 @@ def fix_license(src_lfile, dst_lfile):
     f.close()
 
 
-def check_license(target_config, license_confirmed):
+def check_license(target_config, license_confirmed, introspecting):
     """ Handle the validation of the PyQt5 license.  target_config is the
     target configuration.  license_confirmed is set if the user has already
-    accepted the license.
+    accepted the license.  introspecting is set if the configuration is being
+    determined by introspection.
     """
 
     try:
@@ -2480,7 +2528,7 @@ def check_license(target_config, license_confirmed):
                             sys.platform))
 
     # Common checks.
-    if target_config.qt_licensee != 'Open Source' and ltype == 'GPL':
+    if introspecting and target_config.qt_licensee != 'Open Source' and ltype == 'GPL':
         error(
                 "This version of PyQt5 and the commercial version of Qt have "
                 "incompatible licenses.")
@@ -2671,7 +2719,8 @@ def main(argv):
     target_config.apply_post_options(opts)
 
     # Check the licenses are compatible.
-    check_license(target_config, opts.license_confirmed)
+    check_license(target_config, opts.license_confirmed,
+            (opts.config_file is None))
 
     # Check Python is what we need.
     check_python(target_config)
@@ -2689,6 +2738,8 @@ def main(argv):
     if len(target_config.pyqt_modules) == 0:
         check_modules(target_config, opts.verbose)
 
+    check_dbus(target_config, opts.verbose)
+
     # Tell the user what's been found.
     inform_user(target_config, sip_version)
 
